/********************************************************************************
*                                                                               *
*                  U n d o / R e d o - a b l e   C o m m a n d                  *
*                                                                               *
*********************************************************************************
* Copyright (C) 2000,2003 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: FXUndoList.cpp,v 1.29 2003/09/15 16:37:24 fox Exp $                      *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXWindow.h"
#include "FXUndoList.h"

/*
  Notes:

  - When a command is undone, its moved to the redo list.

  - When a command is redone, its moved back to the undo list.

  - Whenever adding a new command, the redo list is deleted.

  - At any time, you can trim down the undo list down to a given
    maximum size or a given number of undo records.  This should
    keep the memory overhead within sensible bounds.

  - To keep track of when we get back to an "unmodified" state, a mark
    can be set.  The mark is basically a counter which is incremented
    with every undo record added, and decremented when undoing a command.
    When we get back to 0, we are back to the unmodified state.

    If, after setting the mark, we have called undo(), then the mark can be
    reached by calling redo().

    If the marked position is in the redo-list, then adding a new undo
    record will cause the redo-list to be deleted, and the marked position
    will become unreachable.

    The marked state may also become unreachable when the undo list is trimmed.

  - You can call also kill the redo list without adding a new command
    to the undo list, although this may cause the marked position to
    become unreachable.

  - We measure the size of the undo-records in the undo-list; when the
    records are moved to the redo-list, they usually contain different
    information!

  - Because we may need to know during execution of a command whether this
    was due to undoing or directly issued by the user, we keep a flag "working"
    which is TRUE during a undo or redo operation.
    
  - Undo commands may be merged; but no merge should take place if this
    would make the marked position unreachable; thus, if the current
    position is marked, a new undo record must be made; otherwise, undo
    records may be merged.
*/

#define NOMARK 2147483647       // No mark is set

using namespace FX;

/*******************************************************************************/

namespace FX {

// Map
FXDEFMAP(FXUndoList) FXUndoListMap[]={
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_CLEAR,      FXUndoList::onCmdClear),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_CLEAR,      FXUndoList::onUpdClear),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_REVERT,     FXUndoList::onCmdRevert),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_REVERT,     FXUndoList::onUpdRevert),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_UNDO,       FXUndoList::onCmdUndo),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_UNDO,       FXUndoList::onUpdUndo),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_REDO,       FXUndoList::onCmdRedo),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_REDO,       FXUndoList::onUpdRedo),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_UNDO_ALL,   FXUndoList::onCmdUndoAll),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_UNDO_ALL,   FXUndoList::onUpdUndo),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_REDO_ALL,   FXUndoList::onCmdRedoAll),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_REDO_ALL,   FXUndoList::onUpdRedo),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_UNDO_COUNT, FXUndoList::onUpdUndoCount),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_REDO_COUNT, FXUndoList::onUpdRedoCount),
  };


// Object implementation
FXIMPLEMENT(FXUndoList,FXObject,FXUndoListMap,ARRAYNUMBER(FXUndoListMap))


// Default implementation of undo name is just "Undo"
FXString FXCommand::undoName() const { return "Undo"; }


// Default implementation of redo name is just "Redo"
FXString FXCommand::redoName() const { return "Redo"; }


// Default implementation of merge is not to merge
FXbool FXCommand::merge(FXCommand*){ return FALSE; }


// Make new empty undo list
FXUndoList::FXUndoList(){
  redolist=NULL;
  undolist=NULL;
  undocount=0;
  redocount=0;
  marker=NOMARK;
  size=0;
  working=FALSE;
  merge=FALSE;
  }


// Mark current state
void FXUndoList::mark(){
  marker=0;
  }


// Unmark undo list
void FXUndoList::unmark(){
  marker=NOMARK;
  }


// Check if marked
FXbool FXUndoList::marked() const {
  return marker==0;
  }


// Cut the redo list
void FXUndoList::cut(){
  register FXCommand *p;
  while(redolist){
    p=redolist;
    redolist=redolist->next;
    delete p;
    }
  redolist=NULL;
  redocount=0;
  if(marker<0) marker=NOMARK;           // Can not return to marked state anymore!
  }


// Add new command, executing if desired
FXbool FXUndoList::add(FXCommand* command,FXbool doit){
  cut();
  if(command){
    working=TRUE;
    
    // Execute command
    if(doit) command->redo();
    
    // We don't try to merge when currently at the marked position, as we
    // would no longer be able to reach a marked position.
    if(merge && marker && undolist && undolist->merge(command)){
      delete command;                   
      }
      
    // Otherwise append the new command to the undo list
    else{
      command->next=undolist;           // Hang into undolist AFTER redo
      undolist=command;
      size+=command->size();		// Measure AFTER redo!
      undocount++;
      if(marker!=NOMARK) marker++;
      }
    FXTRACE((100,"FXUndoList::add: size=%d undocount=%d marker=%d\n",size,undocount,marker));
    working=FALSE;
    return TRUE;
    }
  return FALSE;
  }


// Undo last command
FXbool FXUndoList::undo(){
  register FXCommand *command;
  if(undolist){
    working=TRUE;
    command=undolist;                   // Remove from undolist BEFORE undo
    undolist=undolist->next;
    size-=command->size();		// Measure BEFORE undo!
    command->undo();
    command->next=redolist;             // Hang into redolist AFTER undo
    redolist=command;
    undocount--;
    redocount++;
    if(marker!=NOMARK) marker--;
    FXTRACE((100,"FXUndoList::undo: size=%d undocount=%d redocount=%d marker=%d\n",size,undocount,redocount,marker));
    working=FALSE;
    return TRUE;
    }
  return FALSE;
  }


// Redo next command
FXbool FXUndoList::redo(){
  register FXCommand *command;
  if(redolist){
    working=TRUE;
    command=redolist;                   // Remove from redolist BEFORE redo
    redolist=redolist->next;
    command->redo();
    size+=command->size();		// Measure AFTER redo!
    command->next=undolist;             // Hang into undolist AFTER redo
    undolist=command;
    undocount++;
    redocount--;
    if(marker!=NOMARK) marker++;
    FXTRACE((100,"FXUndoList::redo: size=%d undocount=%d redocount=%d marker=%d\n",size,undocount,redocount,marker));
    working=FALSE;
    return TRUE;
    }
  return FALSE;
  }


// Undo all commands
FXbool FXUndoList::undoAll(){
  while(canUndo()) undo();
  return TRUE;
  }


// Redo all commands
FXbool FXUndoList::redoAll(){
  while(canRedo()) redo();
  return TRUE;
  }


// Revert to marked
FXbool FXUndoList::revert(){
  if(marker!=NOMARK){
    while(marker>0) undo();
    while(marker<0) redo();
    return TRUE;
    }
  return FALSE;
  }


// Can we undo more commands
FXbool FXUndoList::canUndo() const {
  return undolist!=NULL;
  }


// Can we redo more commands
FXbool FXUndoList::canRedo() const {
  return redolist!=NULL;
  }


// Can revert to marked
FXbool FXUndoList::canRevert() const {
  return marker!=NOMARK && marker!=0;
  }


// Clear list
void FXUndoList::clear(){
  register FXCommand *p;
  FXTRACE((100,"FXUndoList::clear: size=%d undocount=%d redocount=%d marker=%d\n",size,undocount,redocount,marker));
  while(redolist){
    p=redolist;
    redolist=redolist->next;
    delete p;
    }
  while(undolist){
    p=undolist;
    undolist=undolist->next;
    delete p;
    }
  redolist=NULL;
  undolist=NULL;
  marker=NOMARK;
  undocount=0;
  redocount=0;
  size=0;
  }


// Clear undo list
long FXUndoList::onCmdClear(FXObject*,FXSelector,void*){
  clear();
  return 1;
  }


// Update Clear undo list
long FXUndoList::onUpdClear(FXObject* sender,FXSelector,void*){
  sender->handle(this,(canUndo()||canRedo())?FXSEL(SEL_COMMAND,FXWindow::ID_ENABLE):FXSEL(SEL_COMMAND,FXWindow::ID_DISABLE),NULL);
  return 1;
  }


// Revert to marked
long FXUndoList::onCmdRevert(FXObject*,FXSelector,void*){
  revert();
  return 1;
  }


// Update revert to marked
long FXUndoList::onUpdRevert(FXObject* sender,FXSelector,void*){
  sender->handle(this,canRevert()?FXSEL(SEL_COMMAND,FXWindow::ID_ENABLE):FXSEL(SEL_COMMAND,FXWindow::ID_DISABLE),NULL);
  return 1;
  }


// Undo last command
long FXUndoList::onCmdUndo(FXObject*,FXSelector,void*){
  undo();
  return 1;
  }


// Undo all commands
long FXUndoList::onCmdUndoAll(FXObject*,FXSelector,void*){
  undoAll();
  return 1;
  }


// Update undo last command
long FXUndoList::onUpdUndo(FXObject* sender,FXSelector,void*){
  sender->handle(this,canUndo()?FXSEL(SEL_COMMAND,FXWindow::ID_ENABLE):FXSEL(SEL_COMMAND,FXWindow::ID_DISABLE),NULL);
  return 1;
  }


// Redo last command
long FXUndoList::onCmdRedo(FXObject*,FXSelector,void*){
  redo();
  return 1;
  }


// Redo all commands
long FXUndoList::onCmdRedoAll(FXObject*,FXSelector,void*){
  redoAll();
  return 1;
  }


// Update redo last command
long FXUndoList::onUpdRedo(FXObject* sender,FXSelector,void*){
  sender->handle(this,canRedo()?FXSEL(SEL_COMMAND,FXWindow::ID_ENABLE):FXSEL(SEL_COMMAND,FXWindow::ID_DISABLE),NULL);
  return 1;
  }


// Update undo count
long FXUndoList::onUpdUndoCount(FXObject* sender,FXSelector,void*){
  sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_SETINTVALUE),(void*)&undocount);
  return 1;
  }


// Update redo count
long FXUndoList::onUpdRedoCount(FXObject* sender,FXSelector,void*){
  sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_SETINTVALUE),(void*)&redocount);
  return 1;
  }


// Trim undo list down to at most n records
void FXUndoList::trimCount(FXint nc){
  FXTRACE((100,"FXUndoList::trimCount: was: size=%d undocount=%d; marker=%d ",size,undocount,marker));
  if(undocount>nc){
    register FXCommand **pp=&undolist;
    register FXCommand *p=*pp;
    register FXint i=0;
    while(p && i<nc){
      pp=&p->next;
      p=*pp;
      i++;
      }
    while(*pp){
      p=*pp;
      *pp=p->next;
      size-=p->size();
      undocount--;
      delete p;
      }
    if(marker>undocount) marker=NOMARK;
    }
  FXTRACE((100,"now: size=%d undocount=%d; marker=%d\n",size,undocount,marker));
  }


// Trim undo list down to at most size sz
void FXUndoList::trimSize(FXuint sz){
  FXTRACE((100,"FXUndoList::trimSize: was: size=%d undocount=%d; marker=%d ",size,undocount,marker));
  if(size>sz){
    register FXCommand **pp=&undolist;
    register FXCommand *p=*pp;
    register FXuint s=0;
    while(p && (s=s+p->size())<=sz){
      pp=&p->next;
      p=*pp;
      }
    while(*pp){
      p=*pp;
      *pp=p->next;
      size-=p->size();
      undocount--;
      delete p;
      }
    if(marker>undocount) marker=NOMARK;
    }
  FXTRACE((100,"now: size=%d undocount=%d; marker=%d\n",size,undocount,marker));
  }


// Clean up
FXUndoList::~FXUndoList(){
  clear();
  }

}

